/**
 * Copyright (c) 2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.log.parser.api;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileFilter;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.TreeSet;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.petals.commons.log.FlowLogData;
import org.ow2.petals.log.parser.api.exception.FlowInstanceLogFileMissingException;
import org.ow2.petals.log.parser.api.exception.FlowRootStepMissingException;
import org.ow2.petals.log.parser.api.exception.FlowStepFieldInvalidOrMissingException;
import org.ow2.petals.log.parser.api.exception.FlowStepFieldRequiredMissingException;
import org.ow2.petals.log.parser.api.model.Flow;
import org.ow2.petals.log.parser.api.model.FlowId;
import org.ow2.petals.log.parser.api.model.FlowStep;

/**
 * A class used to parse and organize Petals logged data.
 * <p>
 * Since Petals 4.2, MONIT-MSG entries have been removed. They are now MONIT entries created by another logger, with
 * additional fields.
 * </p>
 * <p>
 * Usage: <code><pre>
 * FlowBuilder flowBuilder = new FlowBuilder( dir1, dir2, ... );
 * Set<String> flowIds = flowBuilder.findFlowIds();
 * for( String flowId : flowIds ) {
 *      Flow flow = flowBuilder.parseFlow( flowId );
 *      // Do something with the flow
 * }
 * </pre></code>
 * </p>
 *
 * @author Pierre-Yves Gibello - EBM WebSourcing
 * @author Vincent Zurczak - EBM WebSourcing
 */
public class FlowBuilder {

    private File[] logDirectories;

    private final Logger logger;

    /**
     * Constructor.
     */
    public FlowBuilder() {
        this.logger = Logger.getLogger(FlowBuilder.class.getName());
    }

    /**
     * @return the logDirectories
     */
    public File[] getLogDirectories() {
        return this.logDirectories;
    }

    /**
     * Sets the log directories.
     * 
     * @param logDirectories
     *            a list of directories containing Petals logs (not null, not empty)
     */
    public void setLogDirectories(final File... logDirectories) {

        if (logDirectories == null || logDirectories.length == 0)
            throw new IllegalArgumentException("At least one directory must be passed as parameter.");

        this.logDirectories = logDirectories;
        for (final File dir : this.logDirectories) {
            if (dir == null) {
                throw new IllegalArgumentException("A null directory was passed in the list.");
            }

            if (!dir.isDirectory()) {
                throw new IllegalArgumentException(dir + " is not a valid directory.");
            }
        }
    }

    /**
     * Sets the log directories.
     * 
     * @param logDirectories
     *            a list of directory paths containing Petals logs (not null, not empty)
     */
    public void setLogDirectories(final String... logDirectories) {
        this.setLogDirectories(FlowBuilder.convertToFiles(logDirectories));
    }

    /**
     * Parses a flow directory's content and builds a {@link Flow} instance. Invalid data on MONIT traces are ignored.
     * 
     * @param flowId
     *            a flow ID (not null, not the empty string)
     * @return a flow, or null if it could not be built
     * @throws FlowInstanceLogFileMissingException
     *             The log file of the given flow instance is missing
     * @throws FlowStepFieldRequiredMissingException
     *             A required flow step field is missing
     * @throws FlowRootStepMissingException
     *             No root step found for the given flow instance
     */
    public Flow parseFlow(final String flowId) throws FlowInstanceLogFileMissingException,
            FlowStepFieldRequiredMissingException, FlowRootStepMissingException {
        try {
            return this.parseFlow(flowId, false);
        } catch (final FlowStepFieldInvalidOrMissingException e) {
            // Should not occur because errors are ignored during parsing
            throw new RuntimeException(e);
        }
    }

    /**
     * Parses a flow directory's content and builds a {@link Flow} instance.
     * 
     * @param flowId
     *            a flow ID (not null, not the empty string)
     * @param failOnInvalidData
     *            If <code>true</code>, the exception {@link FlowStepFieldInvalidOrMissingException} is thrown when
     *            detecting invalid data as MONIT fields set to null.
     * @return a flow, or null if it could not be built
     * @throws FlowInstanceLogFileMissingException
     *             The log file of the given flow instance is missing
     * @throws FlowStepFieldInvalidOrMissingException
     *             when a flow step field as an invalid value or is missing
     * @throws FlowStepFieldRequiredMissingException
     *             A required flow step field is missing
     * @throws FlowRootStepMissingException
     *             No root step found for the given flow instance
     */
    public Flow parseFlow(final String flowId, final boolean failOnInvalidData)
            throws FlowInstanceLogFileMissingException, FlowStepFieldInvalidOrMissingException,
            FlowStepFieldRequiredMissingException, FlowRootStepMissingException {

        if (flowId == null || flowId.trim().length() == 0) {
            throw new IllegalArgumentException("A flow ID cannot be null or empty.");
        }

        // Index all the MONIT steps
        final Map<String, FlowStep> flowStepIdToFlowStep = new HashMap<String, FlowStep>();
        final Map<String, String> flowStepIdToPreviousStepId = new HashMap<String, String>();
        final Collection<String> rawMonitEntries = readRawMonitEntries(flowId);

        if (this.logger.isLoggable(Level.FINE)) {
            this.logger.fine("Indexing MONIT steps for " + flowId);
        }

        for (final String rme : rawMonitEntries) {
            final String[] values = rme.split(PetalsLogConstants.SEP);
            if (values.length != 2) {
                continue;
            }

            final Map<String, String> properties = FlowBuilder.parseStep(values[1]);

            // Even if the field 'flowInstanceId' is not needed by this parser because it uses the directory as
            // flowInstanceId, we check that its value is correct
            if (failOnInvalidData) {
                final String flowInstanceId = properties.get(FlowLogData.FLOW_INSTANCE_ID_PROPERTY_NAME);
                if (flowInstanceId == null || "null".equals(flowInstanceId)) {
                    throw new FlowStepFieldInvalidOrMissingException(
                            FlowLogData.FLOW_INSTANCE_ID_PROPERTY_NAME, flowInstanceId);
                }
            }

            final String flowStepId = properties.get(FlowLogData.FLOW_STEP_ID_PROPERTY_NAME);
            if (flowStepId == null) {
                throw new FlowStepFieldRequiredMissingException(FlowLogData.FLOW_STEP_ID_PROPERTY_NAME);
            }
            if (failOnInvalidData && "null".equals(flowStepId)) {
                throw new FlowStepFieldInvalidOrMissingException(FlowLogData.FLOW_STEP_ID_PROPERTY_NAME,
                        flowStepId);
            }
            FlowStep flowStep = flowStepIdToFlowStep.get(flowStepId);
            if (flowStep == null) {
                flowStep = new FlowStep(flowStepId);
                flowStepIdToFlowStep.put(flowStepId, flowStep);
            }

            final String traceCode = properties.get(FlowLogData.FLOW_STEP_TRACE_CODE);
            if (traceCode != null && traceCode.startsWith("consume")) {
                flowStep.setType(FlowStep.CONSUME_FLOW_STEP);
            } else if (traceCode != null && traceCode.startsWith("provide")) {
                flowStep.setType(FlowStep.PROVIDE_FLOW_STEP);
            } else if (failOnInvalidData) {
                throw new FlowStepFieldInvalidOrMissingException(FlowLogData.FLOW_STEP_TRACE_CODE, traceCode);
            }

            flowStep.setLogFile(new File(values[0]));
            final String dumpPath = properties.get(PetalsLogConstants.DUMP_PATH);
            if (traceCode != null && traceCode.endsWith("Begin")) {
                flowStep.setBeginProperties(properties);
                flowStep.getBeginProperties().put(PetalsLogConstants.DUMP_PATH, dumpPath);
            } else {
                flowStep.setOutcomeProperties(properties);
                flowStep.getOutcomeProperties().put(PetalsLogConstants.DUMP_PATH, dumpPath);
                if (traceCode != null && traceCode.endsWith("Failure")) {
                    flowStep.setFailure(true);
                }
            }

            final String previousFlowStepId = properties.get(FlowLogData.FLOW_PREVIOUS_STEP_ID_PROPERTY_NAME);
            if (previousFlowStepId != null) {
                flowStepIdToPreviousStepId.put(flowStepId, previousFlowStepId);
            }
        }

        // Build the dependencies between the flow steps
        // This step was prepared before
        if (this.logger.isLoggable(Level.FINE)) {
            this.logger.fine("Building the flow step dependencies for " + flowId);
        }

        List<FlowStep> rootCandidates = new ArrayList<FlowStep>();
        for (final FlowStep flowStep : flowStepIdToFlowStep.values()) {
            final String previousFlowStepId = flowStepIdToPreviousStepId.get(flowStep.getId());
            FlowStep parent = null;

            if (previousFlowStepId != null && (parent = flowStepIdToFlowStep.get(previousFlowStepId)) != null) {
                parent.addChild(flowStep);
                flowStep.setParent(parent);

            } else {
                // No parent? It may be the root.
                rootCandidates.add(flowStep);
            }
        }

        // If multiple root candidates, filter those with a previous step ID
        // (the real one should have none... or nobody can guess which one it is !)
        if (rootCandidates.size() > 1) {
            final List<FlowStep> realCandidates = new ArrayList<FlowStep>();
            for (final FlowStep step : rootCandidates) {
                final String previousFlowStepId = flowStepIdToPreviousStepId.get(step.getId());
                if (previousFlowStepId == null)
                    realCandidates.add(step);
            }

            rootCandidates = realCandidates;
        }

        // There should be only one root
        final Flow result;
        if (rootCandidates.size() == 1)
            result = new Flow(flowId, rootCandidates.get(0));
        else {
            if (failOnInvalidData) {
                throw new FlowRootStepMissingException(flowId);
            } else {
                this.logger.warning("Could not find the flow root step. Flow build is given up.");
                result = null;
            }
        }

        return result;
    }

    /**
     * <p>
     * The flow IDS are sorted by date (most recent first).
     * </p>
     * 
     * @return a set flow IDs contained in the log directories sorted by decreasing timestamp.
     */
    public Set<FlowId> findFlowIds() {

        if (this.logger.isLoggable(Level.FINE)) {
            this.logger.fine("Listing flow IDs.");
        }

        // Find all the flow information
        final Set<FlowId> sortedFlowIds = new TreeSet<FlowId>();

        if (this.logDirectories != null) {
            for (final File dir : this.logDirectories) {
                final FileFilter filterDirs = new FileFilter() {
                    @Override
                    public boolean accept(final File file) {
                        return file.isDirectory() && !file.getName().startsWith(".");
                    }
                };

                final File[] subdirs = dir.listFiles(filterDirs);
                if (subdirs != null) {
                    for (final File sd : subdirs) {
                        final FlowId flowId = new FlowId();
                        flowId.setName(sd.getName());
                        flowId.setLastModification(sd.lastModified());
                        sortedFlowIds.add(flowId);
                    }
                }
            }
        }

        return sortedFlowIds;
    }

    /**
     * <p>
     * Reads the raw MONIT entries in the petals.log file.
     * </p>
     * <p>
     * Every raw log entry starts with the location of the log file where this entry was read. The file path is
     * separated from the log entry by {@link PetalsLogConstants#SEP}.
     * </p>
     *
     * @param flowId
     *            a flow ID (not null, not the empty string)
     * @return a list of raw log entries, or an empty list if the flow instance does not exist
     */
    private Collection<String> readRawMonitEntries(final String flowId) throws FlowInstanceLogFileMissingException {

        if (flowId == null || flowId.trim().length() == 0) {
            throw new IllegalArgumentException("A flow ID cannot be null or empty.");
        }

        final List<String> rawMonitEntries = new ArrayList<String>();
        for (final File dir : this.logDirectories) {
            final File flowInstanceDir = new File(dir, flowId);
            if (!flowInstanceDir.exists()) {
                // Perhaps the flow instance is in other log directory
                continue;
            }

            final File log = new File(flowInstanceDir, "petals.log");
            if (!log.exists()) {
                throw new FlowInstanceLogFileMissingException(flowId, log);
            }

            try (BufferedReader in = new BufferedReader(new FileReader(log));) {
                String line;
                while ((line = in.readLine()) != null) {
                    if (line.contains(" MONIT ")) {
                        rawMonitEntries.add(log.getAbsolutePath() + PetalsLogConstants.SEP + line);
                    }
                }

            } catch (final IOException e) {
                this.logger.log(Level.WARNING, "An error occurred while reading raw MONIT entries. Entry can be lost. "
                        + e.getMessage());

            }
        }

        return rawMonitEntries;
    }

    /**
     * Parses a step and extracts its properties.
     *
     * @param rawLogEntry
     *            a raw log entry (MONIT). Not <code>null</code>, not empty, the caller is responsible of these checks.
     * @return the step's properties or null if the flow step is invalid
     */
    private static Map<String, String> parseStep(final String rawLogEntry) {

        assert rawLogEntry != null;
        assert !rawLogEntry.trim().isEmpty();

        final Map<String, String> flowStep = new HashMap<String, String>();
        int pos = rawLogEntry.indexOf("] :");
        if (pos > 0) {
            final String header = rawLogEntry.substring(0, pos + 1);
            final String content = rawLogEntry.substring(pos + 4);

            // Parse header (node + date + component)
            StringTokenizer st = new StringTokenizer(header, " \t[]");
            int tok = 1;
            final StringBuilder buf = new StringBuilder();

            while (st.hasMoreTokens()) {
                switch (tok++) {
                    case 1: // Node
                        flowStep.put("node", st.nextToken());
                        break;

                    case 2: // Date
                        buf.append(st.nextToken() + " ");
                        break;

                    case 3: // Time
                        buf.append(st.nextToken() + " ");
                        break;

                    case 4: // GMT+...
                        buf.append(st.nextToken());
                        flowStep.put("date", buf.toString());
                        break;

                    case 5: // MONIT (ignore)
                        st.nextToken();
                        break;

                    case 6:
                        flowStep.put("component", st.nextToken());
                        break;

                    default:
                        st.nextToken();
                        break;
                }
            }

            // Parse the Petals attributes
            st = new StringTokenizer(content, ",");
            while (st.hasMoreTokens()) {
                final String pair = st.nextToken();
                pos = pair.indexOf('=');

                if (pos >= 1) {
                    String val = pair.substring(pos + 1).trim();
                    if (val.startsWith("\'")) {
                        val = val.substring(1);
                    }

                    if (val.endsWith("\'")) {
                        val = val.substring(0, val.length() - 1);
                    }

                    flowStep.put(pair.substring(0, pos - 1).trim(), val);
                }
            }
        }

        return flowStep;
    }

    /**
     * A method used to provide a secondary but still useful constructor.
     * 
     * @param files
     *            an array of file paths (can be null, empty, can contain null values)
     * @return an array of files
     */
    private static File[] convertToFiles(final String[] filePaths) {

        final File[] result;
        if (filePaths != null) {
            result = new File[filePaths.length];
            for (int i = 0; i < filePaths.length; i++)
                result[i] = filePaths[i] == null ? null : new File(filePaths[i]);
        } else {
            result = null;
        }

        return result;
    }
}
