/**
 * Copyright (c) 2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.log.parser.api;

import java.io.File;
import java.util.LinkedHashMap;
import java.util.Map;

import org.ow2.petals.log.parser.api.model.Flow;
import org.ow2.petals.log.parser.api.model.FlowStep;

/**
 * @author Pierre-Yves Gibello - EBM WebSourcing
 * @author Vincent Zurczak - EBM WebSourcing
 */
public class PetalsLogUtils {

    private PetalsLogUtils() {
        // No constructor because it's an utility class
    }

    /**
     * Determines whether a flow step is complete or not.
     * <p>
     * A flow step is complete if there is a "begin" step and an "end" or "failure" step.
     * </p>
     *
     * @param flowStep
     *            a flow step
     * @return true if it is complete, false otherwise
     */
    public static boolean isFlowStepComplete(final FlowStep flowStep) {
        return flowStep.getBeginProperties() != null && flowStep.getOutcomeProperties() != null;
    }

    /**
     * Resolves the absolute location of a dump file.
     * <p>
     * If there is no dump file property set, the result will be null.<br />
     * If the dump file is an absolute location, this function checks that this file exists.<br />
     * If the dump file does not point to an existing file, this function tries to resolve it against the location of
     * the log file (relative location).
     * </p>
     *
     * @param flowStep
     *            a flow step (not null)
     * @param begin
     *            true if we want to look at the begin properties, false for the outcome ones
     * @return an existing file, or null otherwise
     */
    public static File resolveDumpFileLocation(final FlowStep flowStep, final boolean begin) {

        final Map<String, String> map = begin ? flowStep.getBeginProperties() : flowStep.getOutcomeProperties();
        final String dp = map.get(PetalsLogConstants.DUMP_PATH);
        final File logFile = flowStep.getLogFile();

        final File result;
        if (dp != null) {
            File f = new File(dp);
            if (f.exists()) {
                result = f;
            } else if (logFile != null && (f = new File(logFile.getParent(), dp)).exists()) {
                result = f;
            } else {
                result = null;
            }
        } else {
            result = null;
        }

        return result;
    }

    /**
     * Linearize a flow step tree.
     * <p>
     * A flow is simply a tree, with a root.<br />
     * This root is a flow step. A flow step has children. It generally has a parent, except the root.So, a flow is
     * simply a tree of flow steps.
     * </p>
     * <p>
     * This function builds a list of flow steps from such a tree.<br />
     * The tree run starts with the root, then the first child, then the first grand-child... Once all the children of a
     * flow step have been processed, its siblings are treated.
     * </p>
     *
     * @param flowStep
     *            a flow step (not null)
     * @return a non-null map, where the key is the flow step ID and the value is the indentation level
     */
    public static Map<FlowStep, Integer> linearizeFlowTree(final FlowStep flowStep) {

        final Map<FlowStep, Integer> flowStepIdToIndentation = new LinkedHashMap<FlowStep, Integer>();
        FlowStep current = flowStep;
        int level = 0;
        nextOne: while (current != null) {

            // Process the current step
            // Precondition: "current" was not already added in the map.
            flowStepIdToIndentation.put(current, level);

            // Find the next one in the children
            while (current != null) {
                // Precondition: the parent has already been processed
                for (final FlowStep child : current.getChildren()) {
                    if (!flowStepIdToIndentation.containsKey(child)) {
                        current = child;
                        level++;
                        continue nextOne;
                    }
                }

                // The root flow step does not have a parent
                current = current.getParent();
                level--;

                // If we quit this loop with current = null, we will exit the other loop.
            }

            // Postcondition: "current" has not been added in the map.
        }

        return flowStepIdToIndentation;
    }

    /**
     * Finds a flow step.
     * 
     * @param flow
     *            a flow (not null)
     * @param stepId
     *            the step ID
     * @return a flow step (may be null)
     */
    public static FlowStep findFlowStep(final Flow flow, final String stepId) {

        final FlowStep stepObject;
        if (stepId.equals(flow.getRoot().getId())) {
            stepObject = flow.getRoot();
        } else {
            stepObject = flow.getRoot().findDescendant(stepId);
        }

        return stepObject;
    }
}