/**
 * Copyright (c) 2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.log.parser.api.model;

import java.io.File;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import org.ow2.petals.log.parser.api.PetalsLogConstants;

/**
 * A flow step represents a set of log entries with the MONIT or MONIT-MSG level.
 * <p>
 * Every flow step has a unique ID.<br />
 * It matches an exchange in Petals (request / response / ACK).<br />
 * It has a "begin" part and an "end" or "failure" part.
 * </p>
 * <p>
 * A flow step may have a parent and children too.<br />
 * This relation is deduced from the {@link PetalsLogConstants#PROVIDE_FLOW_PREVIOUS_STEP_ID} property.
 * </p>
 *
 * @author Pierre-Yves Gibello - EBM WebSourcing
 */
public class FlowStep implements Comparable<FlowStep> {

    public static final int PROVIDE_FLOW_STEP = 0;

    public static final int CONSUME_FLOW_STEP = 1;

    /**
     * The flow step ID.
     */
    private final String flowStepId;

    /**
     * The type of the flow step.
     * <p>
     * Default is {@link #PROVIDE_FLOW_STEP}.
     * </p>
     */
    private int flowStepType = PROVIDE_FLOW_STEP;

    /**
     * The properties of the BEGIN.
     */
    private Map<String, String> propertiesOnFlowStepBegin;

    /**
     * The properties of the END or FAULT.
     */
    private Map<String, String> propertiesOnFlowStepEnd;

    /**
     * Indicates whether the outcome is a FAULT or an END.
     * <p>
     * Default is <b>false</b>, ie we assume it is an END.
     * </p>
     */
    private boolean isFailure;

    /**
     * The children steps (Petals exchanges that were created because of the current exchange).
     */
    private final SortedSet<FlowStep> flowStepChildren = new TreeSet<FlowStep>();

    /**
     * A map indexing the children, so that search goes faster.
     */
    private Map<String, FlowStep> childIdToChildStep = new HashMap<String, FlowStep>();

    /**
     * The parent of this flow step.
     */
    private FlowStep parent;

    /**
     * The log file, useful to find the absolute path of a dumped payload.
     */
    private File logFile;

    /**
     * Constructor.
     * 
     * @param id
     *            the flow step's ID. Not <code>null</code>
     */
    public FlowStep(final String id) {
        assert id != null;

        this.flowStepId = id;
    }

    /**
     * @return the id
     */
    public String getId() {
        return this.flowStepId;
    }

    /**
     * @return the type_
     */
    public int getType() {
        return this.flowStepType;
    }

    /**
     * @param type
     *            the type_ to set
     */
    public void setType(final int type) {
        this.flowStepType = type;
    }

    /**
     * @return the begin
     */
    public Map<String, String> getBeginProperties() {
        return this.propertiesOnFlowStepBegin;
    }

    /**
     * @param begin
     *            the begin to set
     */
    public void setBeginProperties(final Map<String, String> begin) {
        this.propertiesOnFlowStepBegin = begin;
    }

    /**
     * @return the outcome_
     */
    public Map<String, String> getOutcomeProperties() {
        return this.propertiesOnFlowStepEnd;
    }

    /**
     * @param outcome_
     *            the outcome_ to set
     */
    public void setOutcomeProperties(final Map<String, String> outcome) {
        this.propertiesOnFlowStepEnd = outcome;
    }

    /**
     * @return the failure_
     */
    public boolean isFailure() {
        return this.isFailure;
    }

    /**
     * @param failure
     *            the failure_ to set
     */
    public void setFailure(final boolean failure) {
        this.isFailure = failure;
    }

    /**
     * @return the parent
     */
    public FlowStep getParent() {
        return this.parent;
    }

    /**
     * @return the children
     */
    public Set<FlowStep> getChildren() {
        return this.flowStepChildren;
    }

    /**
     * @return the number of children
     */
    public int getChildCount() {
        return this.flowStepChildren.size();
    }

    /**
     * Finds a descendant step.
     * 
     * @param flowStepId
     *            the flow step ID
     * @return the flow step, or null if it was not found
     */
    public FlowStep findDescendant(final String flowStepId) {

        FlowStep child = this.childIdToChildStep.get(flowStepId);
        if (child == null) {
            for (final FlowStep step : this.flowStepChildren) {
                child = step.findDescendant(flowStepId);
                if (child != null)
                    break;
            }
        }

        return child;
    }

    /**
     * @return the IDs of the direct children steps
     */
    public Set<String> getChildrenFlowStepIds() {
        return this.childIdToChildStep.keySet();
    }

    /**
     * Sets the parent
     * 
     * @param parent
     */
    public void setParent(final FlowStep parent) {
        this.parent = parent;
    }

    /**
     * Adds a child.
     * 
     * @param child
     */
    public void addChild(final FlowStep child) {
        this.flowStepChildren.add(child);
        this.childIdToChildStep.put(child.getId(), child);
    }

    /**
     * @return the logFile_
     */
    public File getLogFile() {
        return this.logFile;
    }

    /**
     * @param logFile
     *            the logFile to set
     */
    public void setLogFile(final File logFile) {
        this.logFile = logFile;
    }

    /**
     * Prints the flow step with its children.
     * 
     * @param buf
     *            a string buffer (not null)
     * @param level
     *            the level (0 for the tree root)
     * @return a string representation of the flow step and its descendants
     */
    protected String toString(final StringBuilder buf, final int level) {

        if (level > 0) {
            buf.append('|');
        }

        for (int i = 0; i < level; i++) {
            buf.append("-");
        }

        buf.append(" " + this.flowStepId + "\n");
        for (final FlowStep c : this.flowStepChildren) {
            c.toString(buf, level + 1);
        }

        return buf.toString();
    }

    @Override
    public String toString() {
        final StringBuilder strBuf = new StringBuilder(this.flowStepId);
        if (this.propertiesOnFlowStepBegin != null) {
            strBuf.append(": ");
            for (final Entry<String, String> property : this.propertiesOnFlowStepBegin.entrySet()) {
                strBuf.append(property.getKey());
                strBuf.append("=");
                strBuf.append('\'').append(property.getValue()).append("',");
            }
        }
        return strBuf.toString();
    }

    @Override
    public int compareTo(final FlowStep flowStep) {
        final String thisDate = this.propertiesOnFlowStepBegin.get(PetalsLogConstants.DATE);
        return thisDate.compareTo(flowStep.getBeginProperties().get(PetalsLogConstants.DATE));
    }

    @Override
    public int hashCode() {
        assert this.flowStepId != null;
        return 31 + this.flowStepId.hashCode();
    }

    @Override
    public boolean equals(final Object obj) {
        if (this == obj)
            return true;
        if (obj == null)
            return false;
        if (getClass() != obj.getClass())
            return false;
        FlowStep other = (FlowStep) obj;
        assert this.flowStepId != null;
        assert other.flowStepId != null;
        return this.flowStepId.equals(other.flowStepId);
    }
}
