/**
 * Copyright (c) 2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.log.parser.api;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.io.File;
import java.io.IOException;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;

import org.junit.Test;
import org.ow2.petals.log.parser.api.exception.FlowInstanceLogFileMissingException;
import org.ow2.petals.log.parser.api.exception.FlowRootStepMissingException;
import org.ow2.petals.log.parser.api.exception.FlowStepFieldInvalidOrMissingException;
import org.ow2.petals.log.parser.api.exception.FlowStepFieldRequiredMissingException;
import org.ow2.petals.log.parser.api.miscellaneous.TestUtils;
import org.ow2.petals.log.parser.api.model.Flow;
import org.ow2.petals.log.parser.api.model.FlowId;
import org.ow2.petals.log.parser.api.model.FlowStep;

/**
 * @author Vincent Zurczak - EBM WebSourcing
 * @author Christophe DENEUX - Linagora
 */
public class FlowBuilderTestCase {

    /**
     * <p>
     * Performs tests on {@link FlowBuilder} with one node logs, where a flow step contains an invalid value for
     * traceCode.
     * </p>
     * <p>
     * Note: The parsing is done with and without ignoring errors.
     * </p>
     */
    @Test
    public void parseLog_TraceCodeInvalid() throws IOException, URISyntaxException, FlowInstanceLogFileMissingException,
            FlowStepFieldInvalidOrMissingException, FlowStepFieldRequiredMissingException,
            FlowRootStepMissingException {
        this.readInvalidLogFiles("traceCodeInvalid");
    }

    private void readInvalidLogFiles(final String logFile) throws IOException, URISyntaxException,
            FlowInstanceLogFileMissingException, FlowStepFieldInvalidOrMissingException,
            FlowStepFieldRequiredMissingException, FlowRootStepMissingException {

        // Testing the number of flow IDs
        final File[] rootFiles = TestUtils.findTestDirectories(logFile);
        final FlowBuilder flowBuilder = new FlowBuilder();
        flowBuilder.setLogDirectories(rootFiles);
        final Set<FlowId> flowIds = flowBuilder.findFlowIds();
        assertEquals(1, flowIds.size());
        assertEquals("4332dd30-a72d-11e2-8071-001c231aad1f", flowIds.iterator().next().getName());

        // Ignoring errors
        final Flow flowWithDefaultParsing = flowBuilder.parseFlow(flowIds.iterator().next().getName());
        assertNotNull(flowWithDefaultParsing);
        final Flow flowWithoutError = flowBuilder.parseFlow(flowIds.iterator().next().getName(), false);
        assertNotNull(flowWithoutError);

        // Managing errors
        try {
            flowBuilder.parseFlow(flowIds.iterator().next().getName(), true);
            fail("invalid flow step field not detected");
        } catch (final FlowStepFieldInvalidOrMissingException e) {
            // Expected exception
        }
    }

    /**
     * <p>
     * Performs tests on {@link FlowBuilder} with one node logs, where the trace code is missing on a flow step.
     * </p>
     * <p>
     * Note: The parsing is done with and without ignoring errors.
     * </p>
     */
    @Test
    public void parseLog_TraceCodeMissing() throws IOException, URISyntaxException, FlowInstanceLogFileMissingException,
            FlowStepFieldInvalidOrMissingException, FlowStepFieldRequiredMissingException,
            FlowRootStepMissingException {
        this.readInvalidLogFiles("traceCodeMissing");
    }

    /**
     * <p>
     * Performs tests on {@link FlowBuilder} with one node logs, where a flow step contains an invalid value ("null")
     * for flowInstanceId.
     * </p>
     * <p>
     * Note: The parsing is done with and without ignoring errors.
     * </p>
     */
    @Test
    public void parseLog_FlowInstanceIdInvalid() throws IOException, URISyntaxException,
            FlowInstanceLogFileMissingException, FlowStepFieldInvalidOrMissingException,
            FlowStepFieldRequiredMissingException, FlowRootStepMissingException {
        this.readInvalidLogFiles("flowInstanceIdInvalid");
    }

    /**
     * <p>
     * Performs tests on {@link FlowBuilder} with one node logs, where the flow instance identifier is missing on a flow
     * step.
     * </p>
     * <p>
     * Note: The parsing is done with and without ignoring errors.
     * </p>
     */
    @Test
    public void parseLog_FlowInstanceIdMissing() throws IOException, URISyntaxException,
            FlowInstanceLogFileMissingException, FlowStepFieldInvalidOrMissingException,
            FlowStepFieldRequiredMissingException, FlowRootStepMissingException {
        this.readInvalidLogFiles("flowInstanceIdMissing");
    }

    /**
     * <p>
     * Performs tests on {@link FlowBuilder} with one node logs, where a flow step contains an invalid value ("null")
     * for flowStepId.
     * </p>
     * <p>
     * Note: The parsing is done with and without ignoring errors.
     * </p>
     */
    @Test
    public void parseLog_FlowStepIdInvalid() throws IOException, URISyntaxException,
            FlowInstanceLogFileMissingException, FlowStepFieldInvalidOrMissingException,
            FlowStepFieldRequiredMissingException, FlowRootStepMissingException {

        // Testing the number of flow IDs
        final File[] rootFiles = TestUtils.findTestDirectories("flowStepIdInvalid");
        final FlowBuilder flowBuilder = new FlowBuilder();
        flowBuilder.setLogDirectories(rootFiles);
        final Set<FlowId> flowIds = flowBuilder.findFlowIds();
        assertEquals(1, flowIds.size());
        assertEquals("4332dd30-a72d-11e2-8071-001c231aad1f", flowIds.iterator().next().getName());

        // Ignoring errors
        flowBuilder.parseFlow(flowIds.iterator().next().getName());
        flowBuilder.parseFlow(flowIds.iterator().next().getName(), false);

        // Managing errors
        try {
            flowBuilder.parseFlow(flowIds.iterator().next().getName(), true);
            fail("invalid flow step field not detected");
        } catch (final FlowStepFieldInvalidOrMissingException e) {
            // Expected exception
        }
    }

    /**
     * <p>
     * Performs tests on {@link FlowBuilder} with one node logs, where the flow step identifier is missing on a flow
     * step.
     * </p>
     * <p>
     * Note: The parsing is done with and without ignoring errors.
     * </p>
     */
    @Test
    public void parseLog_FlowStepIdMissing() throws IOException, URISyntaxException,
            FlowInstanceLogFileMissingException, FlowStepFieldInvalidOrMissingException,
            FlowStepFieldRequiredMissingException, FlowRootStepMissingException {

        // Testing the number of flow IDs
        final File[] rootFiles = TestUtils.findTestDirectories("flowStepIdMissing");
        final FlowBuilder flowBuilder = new FlowBuilder();
        flowBuilder.setLogDirectories(rootFiles);
        final Set<FlowId> flowIds = flowBuilder.findFlowIds();
        assertEquals(1, flowIds.size());
        assertEquals("4332dd30-a72d-11e2-8071-001c231aad1f", flowIds.iterator().next().getName());

        // Ignoring errors
        try {
            flowBuilder.parseFlow(flowIds.iterator().next().getName());
            fail("invalid flow step field not detected");
        } catch (final FlowStepFieldRequiredMissingException e) {
            // Expected exception
        }
        try {
            flowBuilder.parseFlow(flowIds.iterator().next().getName(), false);
            fail("invalid flow step field not detected");
        } catch (final FlowStepFieldRequiredMissingException e) {
            // Expected exception
        }

        // Managing errors
        try {
            flowBuilder.parseFlow(flowIds.iterator().next().getName(), true);
            fail("invalid flow step field not detected");
        } catch (final FlowStepFieldRequiredMissingException e) {
            // Expected exception
        }
    }

    /**
     * <p>
     * Performs tests on {@link FlowBuilder} with one node logs, where the root step of flow instance is not found.
     * </p>
     * <p>
     * Note: The parsing is done with and without ignoring errors.
     * </p>
     */
    @Test
    public void parseLog_FlowRootStepMissing() throws IOException, URISyntaxException,
            FlowInstanceLogFileMissingException, FlowStepFieldInvalidOrMissingException,
            FlowStepFieldRequiredMissingException, FlowRootStepMissingException {

        // Testing the number of flow IDs
        final File[] rootFiles = TestUtils.findTestDirectories("flowRootStepMissing");
        final FlowBuilder flowBuilder = new FlowBuilder();
        flowBuilder.setLogDirectories(rootFiles);
        final Set<FlowId> flowIds = flowBuilder.findFlowIds();
        assertEquals(1, flowIds.size());
        assertEquals("1cb1ccf0-eab6-11e5-9eec-024278261286", flowIds.iterator().next().getName());

        // Ignoring errors: the invalid flow instance is skipped
        assertNull(flowBuilder.parseFlow(flowIds.iterator().next().getName()));
        assertNull(flowBuilder.parseFlow(flowIds.iterator().next().getName(), false));

        // Managing errors
        try {
            flowBuilder.parseFlow(flowIds.iterator().next().getName(), true);
            fail("invalid flow step field not detected");
        } catch (final FlowRootStepMissingException e) {
            // Expected exception
        }
    }

    /**
     * Performs tests on {@link FlowBuilder} with one node logs (without dumps).
     */
    @Test
    public void parseLog_SingleNodeWithoutDump() throws Exception {

        // Testing the number of flow IDs
        File[] rootFiles = TestUtils.findTestDirectories("oneNodeWithoutDump");
        FlowBuilder flowBuilder = new FlowBuilder();
        flowBuilder.setLogDirectories(rootFiles);
        Set<FlowId> flowIds = flowBuilder.findFlowIds();
        assertEquals(1, flowIds.size());
        assertEquals("4332dd30-a72d-11e2-8071-001c231aad1f", flowIds.iterator().next().getName());

        // Checking assertions on the final structure
        Flow flow = flowBuilder.parseFlow(flowIds.iterator().next().getName());
        assertNotNull(flow);

        // The root
        assertNotNull(flow.getRoot());
        assertEquals("4332dd31-a72d-11e2-8071-001c231aad1f", flow.getRoot().getId());
        assertNotNull(flow.getRoot().getBeginProperties());
        assertNotNull(flow.getRoot().getOutcomeProperties());
        assertEquals(1, flow.getRoot().getChildren().size());

        // Expected number of flow steps
        assertEquals(2, this.getAllFlowSteps(flow.getRoot()).size());

        assertEquals(flow.getRoot().getType(), FlowStep.CONSUME_FLOW_STEP);
        assertEquals(12, flow.getRoot().getBeginProperties().size());
        assertEquals("http://10.0.0.1:8084/petals/services/Echo",
                flow.getRoot().getBeginProperties().get(PetalsLogConstants.SOAP_CONSUME_REQUESTED_URL));

        assertFalse(flow.getRoot().isFailure());
        assertEquals(7, flow.getRoot().getOutcomeProperties().size());

        // The first (and single) child
        FlowStep firstChild = flow.getRoot().getChildren().iterator().next();
        assertEquals(flow.getRoot(), firstChild.getParent());
        assertEquals(0, firstChild.getChildren().size());
        assertNotNull(firstChild.getBeginProperties());
        assertNotNull(firstChild.getOutcomeProperties());
        assertFalse(firstChild.isFailure());
    }

    /**
     * Performs tests on {@link FlowBuilder} with one node logs (with dumps).
     */
    @Test
    public void parseLog_SingleNodeWithDump() throws Exception {

        // Testing the number of flow IDs
        File[] rootFiles = TestUtils.findTestDirectories("oneNodeWithDump");
        FlowBuilder flowBuilder = new FlowBuilder();
        flowBuilder.setLogDirectories(rootFiles);
        Set<FlowId> flowIds = flowBuilder.findFlowIds();
        assertEquals(1, flowIds.size());

        // instead of using assertEquals on 'FlowId.getName', we prefer an assertEquals on 'flowId' to validate
        // 'FlowId.equals(...)'
        final FlowId expectedFlowId = new FlowId();
        expectedFlowId.setName("9ed1b7c0-a72c-11e2-bb5b-001c231aad1f");
        assertEquals(expectedFlowId, flowIds.iterator().next());

        // Checking assertions on the final structure
        Flow flow = flowBuilder.parseFlow(flowIds.iterator().next().getName());
        assertNotNull(flow);

        // instead of using assertEquals on 'flow.getFlowId', we prefer an assertEquals on 'flow' to validate
        // 'Flow.equals(...)'
        final Flow expectedFlow = new Flow("9ed1b7c0-a72c-11e2-bb5b-001c231aad1f", new FlowStep("a-flow-step-id"));
        assertEquals(expectedFlow, flow);

        // The root
        assertNotNull(flow.getRoot());
        assertEquals("9ed1b7c1-a72c-11e2-bb5b-001c231aad1f", flow.getRoot().getId());
        assertNotNull(flow.getRoot().getBeginProperties());
        assertNotNull(flow.getRoot().getOutcomeProperties());
        assertEquals(1, flow.getRoot().getChildren().size());

        // Expected number of flow steps
        assertEquals(2, this.getAllFlowSteps(flow.getRoot()).size());

        assertEquals(flow.getRoot().getType(), FlowStep.CONSUME_FLOW_STEP);
        assertEquals(12, flow.getRoot().getBeginProperties().size());
        assertEquals("http://10.0.0.1:8084/petals/services/Echo",
                flow.getRoot().getBeginProperties().get(PetalsLogConstants.SOAP_CONSUME_REQUESTED_URL));

        assertTrue(flow.getRoot().isFailure());
        assertEquals(8, flow.getRoot().getOutcomeProperties().size());

        // The first (and single) child
        FlowStep firstChild = flow.getRoot().getChildren().iterator().next();
        assertEquals(flow.getRoot(), firstChild.getParent());
        assertNotNull(firstChild.getBeginProperties());
        assertNotNull(firstChild.getOutcomeProperties());
        assertTrue(firstChild.isFailure());

        String dumpPath = firstChild.getOutcomeProperties().get(PetalsLogConstants.DUMP_PATH);
        assertNotNull(dumpPath);
        File dumpFile = PetalsLogUtils.resolveDumpFileLocation(firstChild, false);
        assertNotNull(dumpFile);

        final FlowStep expectedFlowStep = new FlowStep("9f00de10-a72c-11e2-bb5b-001c231aad1f");
        assertEquals(expectedFlowStep, firstChild);
    }

    /**
     * <p>
     * Performs tests on {@link FlowBuilder} with one node logs (with dumps, without the log file (ie. petals.log or
     * equivalent)).
     * </p>
     * <p>
     * Note: this test uses {@link FlowBuilder#setLogDirectories(String...)} to define the log directories.
     * </p>
     */
    @Test(expected = FlowInstanceLogFileMissingException.class)
    public void parseLog_SingleNodeWithDumpWithoutLogFile() throws IOException, URISyntaxException,
            FlowInstanceLogFileMissingException, FlowStepFieldRequiredMissingException, FlowRootStepMissingException {

        // Testing the number of flow IDs
        final URL rootDirUrl = Thread.currentThread().getContextClassLoader()
                .getResource("oneNodeWithDumpWithoutLogFile/node1");
        assertNotNull(rootDirUrl);
        final String[] rootFiles = new String[] { new File(rootDirUrl.toURI()).getAbsolutePath() };
        final FlowBuilder flowBuilder = new FlowBuilder();
        flowBuilder.setLogDirectories(rootFiles);
        final Set<FlowId> flowIds = flowBuilder.findFlowIds();
        assertEquals(1, flowIds.size());
        assertEquals("9ed1b7c0-a72c-11e2-bb5b-001c231aad1f", flowIds.iterator().next().getName());

        flowBuilder.parseFlow(flowIds.iterator().next().getName());
    }

    /**
     * Test flow sorting by date (most recent first).
     * 
     * @throws Exception
     */
    @Test
    public void testFlowSorting() throws Exception {

        // Testing the number of flow IDs
        final File[] rootFiles = TestUtils.findTestDirectories("severalflows");
        final FlowBuilder flowBuilder = new FlowBuilder();
        flowBuilder.setLogDirectories(rootFiles);

        final Set<FlowId> flowIds = flowBuilder.findFlowIds();
        Long lastModified = 0L;
        String lastName = "";
        for (final FlowId flowId : flowIds) {
            System.out.println(flowId.getName());
            final long lastModification = flowId.getLastModification();
            assertTrue(lastModification + " !> " + lastModified, lastModification >= lastModified);
            if (lastModification == lastModified) {
                assertTrue(flowId.getName().compareTo(lastName) > 0);
            }
            lastModified = lastModification;
            lastName = flowId.getName();
        }
    }

    /**
     * <p>
     * Check {@link FlowBuilder#findFlowIds()} when no log directory was previously set.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>no error,</li>
     * <li>an empty flow instance list is returned.</li>
     * </ul>
     * </p>
     * 
     */
    @Test
    public void findFlowIds_noLogDirectorySet() throws Exception {

        final FlowBuilder flowBuilder = new FlowBuilder();

        final Set<FlowId> flowIds = flowBuilder.findFlowIds();
        assertNotNull(flowIds);
        assertEquals(0, flowIds.size());
    }

    /**
     * <p>
     * Check {@link FlowBuilder#findFlowIds()} when no subdirs exist into the log directories was previously set.
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>no error,</li>
     * <li>an empty flow instance list is returned.</li>
     * </ul>
     * </p>
     * 
     */
    @Test
    public void findFlowIds_logDirectoryWithoutFlowDirs() throws Exception {

        final URL rootDirUrl = Thread.currentThread().getContextClassLoader()
                .getResource("oneNodeWithDump/node1/9ed1b7c0-a72c-11e2-bb5b-001c231aad1f");

        final FlowBuilder flowBuilder = new FlowBuilder();
        flowBuilder.setLogDirectories(new File(rootDirUrl.toURI()));

        final Set<FlowId> flowIds = flowBuilder.findFlowIds();
        assertNotNull(flowIds);
        assertEquals(0, flowIds.size());
    }

    /**
     * <p>
     * Check {@link FlowBuilder#findFlowIds()} when a given log directory is not a directory.
     * </p>
     * <p>
     * Expected results: It is not possible to set a such log directory. An error occurs on
     * {@link FlowBuilder#setLogDirectories(File...)}.
     * </p>
     * 
     */
    @Test
    public void findFlowIds_logDirectoryIsNotADirectory() throws Exception {

        final URL rootDirUrl = Thread.currentThread().getContextClassLoader()
                .getResource("oneNodeWithDump/node1/9ed1b7c0-a72c-11e2-bb5b-001c231aad1f/petals.log");

        final FlowBuilder flowBuilder = new FlowBuilder();
        try {
            flowBuilder.setLogDirectories(new File(rootDirUrl.toURI()));
        } catch (final IllegalArgumentException e) {
            assertTrue(e.getMessage().contains("not a valid directory"));
        }
    }

    /**
     * Build a flow step list ordered by arrival, from a given flow step
     * 
     * @param flowStep
     *            The root flow step from which the list must be build.
     * @return
     */
    private List<FlowStep> getAllFlowSteps(final FlowStep flowStep) {
        final List<FlowStep> allFlowSteps = new ArrayList<FlowStep>();
        allFlowSteps.add(flowStep);
        for (final FlowStep child : flowStep.getChildren()) {
            allFlowSteps.addAll(this.getAllFlowSteps(child));
        }
        return allFlowSteps;
    }
}
